/* $Id: db.c,v 1.66 2007-06-05 16:02:52 kristaps Exp $ */
/*
Copyright (c) 2007 Kristaps Dzonsons
Latvijas Universitaates Matemaatikas un Informaatikas Instituuts
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

* Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
* Neither the name of the copyright holders nor the names of
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <arpa/inet.h>
#include <sys/param.h>

#include <assert.h>
#include <err.h>
#include <math.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include <db.h>

#include "nlpcrawl.h"


/*
 * Maximum window size exponent.
 */
#define DB_SCNPWR_MAX	(31)

/*
 * Minimum window size exponent.
 */
#define DB_SCNPWR_MIN	(0)

/*
 * Default (beginning) window size exponent. 
 */
#define DB_DEF_SCANPWR	(11)

/*
 * Default database-locking timeout.
 */
#define DB_DEF_TIMEO	(1000 * 100)

/*
 * Default memory pool cache size.  Set to 128 MB, which seems
 * reasonable on any machine.  This will, obviously, scale larger.
 */
#define DB_DEF_CACHE	(128 * 1024 * 1024)

#define DB_NO_ERROR	(0)

/*
 * Host to network long-long.
 */
#define _htonll(x) \
	(((uint64_t)htonl((x)) << 32) + htonl((x) >> 32))

/*
 * Network to host long-long.
 */
#define _ntohll(x) \
	(((uint64_t)ntohl(x) << 32) + ntohl(x >> 32))


static struct db_state_t
{
	DB		*dbp;
	DB_ENV		*dbenv;
	char		dbfile[MAXPATHLEN];
	char		dbenvdir[MAXPATHLEN];
} db_state;

static uint8_t		db_scnpwr = DB_DEF_SCANPWR;


static int32_t
db_window_get(void)
{
	uint64_t	t;

	assert(DB_SCNPWR_MAX >= db_scnpwr);
	t = (uint64_t)pow(2, db_scnpwr);
	assert(t <= INT32_MAX);
	return((int32_t)t);
}


void
db_window_grow(void)
{
	db_scnpwr += db_scnpwr == DB_SCNPWR_MAX ? 0 : 1;
	dddbg("%s: window exponent: %d", __func__, db_scnpwr);
}


void
db_window_shrink(void)
{
	db_scnpwr -= db_scnpwr == DB_SCNPWR_MIN ? 0 : 1;
	dddbg("%s: window exponent: %d", __func__, db_scnpwr);
}


void
db_window_fill(struct db_query *p)
{
	assert(NULL != p);
	p->window = db_window_get();
}


/*
 * Copy a host-byte-order structure into network-byte-order.  This must
 * be modified to handle any new structure elements!
 */
static void
copyin(const struct nlpcdbr *in, struct nlpcdbr *out)
{
	assert(NULL != in);
	assert(NULL != out);

	memcpy(out->charset, in->charset, NLPCR_CHARSETLEN);
	memcpy(out->lang, in->lang, NLPCR_LANGLEN);
	memcpy(out->data, in->data, NLPCR_HASHLEN);
	memcpy(out->ip, in->ip, NLPCR_IP_WIDTH);
	memcpy(&out->uri, &in->uri, sizeof(struct uri));

	out->size = _htonll(in->size);
	out->code = _htonll(in->code);
	out->state = _htonll(in->state);
	out->type = _htonll(in->type);
	out->expires = _htonll(in->expires);
	out->ftime = _htonll(in->ftime);
	out->ctime = _htonll(in->ctime);
	out->atime = _htonll(in->atime);
	out->mtime = _htonll(in->mtime);
	out->revived = _htonll(in->revived);
}


/*
 * Copy a network-byte-order structure into host-byte-order.  This must
 * be modified to handle any new structure elements!
 */
static void
copyout(const struct nlpcdbr *in, struct nlpcdbr *out)
{
	assert(NULL != in);
	assert(NULL != out);

	memcpy(out->charset, in->charset, NLPCR_CHARSETLEN);
	memcpy(out->lang, in->lang, NLPCR_LANGLEN);
	memcpy(out->data, in->data, NLPCR_HASHLEN);
	memcpy(out->ip, in->ip, NLPCR_IP_WIDTH);
	memcpy(&out->uri, &in->uri, sizeof(struct uri));

	out->size = _ntohll(in->size);
	out->code = _ntohll(in->code);
	out->state = _ntohll(in->state);
	out->type = _ntohll(in->type);
	out->expires = _ntohll(in->expires);
	out->ftime = _ntohll(in->ftime);
	out->ctime = _ntohll(in->ctime);
	out->atime = _ntohll(in->atime);
	out->mtime = _ntohll(in->mtime);
	out->revived = _ntohll(in->revived);
}


void
db_init(const char *prefix, uint32_t flags)
{
	int 		rc;
	uint32_t	dbflags;
	DB		*dbp;
	DB_ENV		*dbenv;
	char		*dbfile, *dbedir, tmp[MAXPATHLEN];

	assert(NULL != prefix);

	assert(NULL == db_state.dbp);
	assert(NULL == db_state.dbenv);
	assert(0 == db_state.dbfile[0]);
	assert(0 == db_state.dbenvdir[0]);

	dbfile = db_state.dbfile;
	dbedir = db_state.dbenvdir;

	/*
	 * First configure, then construct, our file-system hierarchy.
	 */

	memset(dbfile, 0, MAXPATHLEN);
	memset(dbedir, 0, MAXPATHLEN);

	if (0 == strlen(prefix))
		diex("fail: prefix is zero-length");

	strncat(dbedir, prefix, MAXPATHLEN);
	strncat(dbfile, prefix, MAXPATHLEN);

	if (0 != dbedir[MAXPATHLEN - 20])
		diex("fail: prefix is too long: %s", prefix);

	if ('/' != dbedir[strlen(dbedir) - 1]) {
		strncat(dbedir, "/", MAXPATHLEN - strlen(dbedir));
		assert(0 == dbedir[MAXPATHLEN - 1]);
		strncat(dbfile, "/", MAXPATHLEN - strlen(dbfile));
	}

	strncat(dbedir, "data", MAXPATHLEN);
	strncat(dbfile, "db", MAXPATHLEN);

	if (0 != dbedir[MAXPATHLEN - 1])
		diex("fail: database path too long");
	if (0 != dbfile[MAXPATHLEN - 1])
		diex("fail: database path too long");

	/* 
	 * Note that mkdir_p modifies the underlying buffer, so use a
	 * copy in order to construct the hierarchy. 
	 */ 

	memcpy(tmp, dbfile, MAXPATHLEN);
	if (mkdir_p(tmp, 0755))
		die("fail: mkdir_p: %s", dbfile);
	memcpy(tmp, dbedir, MAXPATHLEN);
	if (mkdir_p(tmp, 0755))
		die("fail: mkdir_p: %s", dbedir);

	strncat(dbfile, "/nlpcrawl.db", MAXPATHLEN);
	if (0 != dbfile[MAXPATHLEN - 1])
		diex("fail: database filename too long");

	dddbg("%s: environment: %s", __func__, dbedir);
	dddbg("%s: file: %s", __func__, dbfile);

	/*
	 * Now create database systems.  Start with the environment.
	 */

	if (0 != (rc = db_env_create(&db_state.dbenv, 0)))
		diex("fail: db_env_create: %s", db_strerror(rc));
	dbenv = db_state.dbenv;

	if (0 != (rc = dbenv->set_timeout(dbenv, DB_DEF_TIMEO, 
					DB_SET_LOCK_TIMEOUT)))
		diex("fail: dbenv set_timeout: %s, %s",
				db_strerror(rc), dbedir);
	if (0 != (rc = dbenv->set_lk_detect(dbenv, DB_LOCK_DEFAULT)))
		diex("fail: dbenv set_lk_detect: %s, %s",
				db_strerror(rc), dbedir);
	if (0 != (rc = dbenv->set_cachesize(dbenv, 0, 
					DB_DEF_CACHE, 0)))
		diex("fail: dbenv cachesize: %s, %s",
				db_strerror(rc), dbedir);
	if (0 != (rc = dbenv->set_data_dir(dbenv, dbedir)))
		diex("fail: dbenv data dir: %s, %s",
				db_strerror(rc), dbedir);
	if (0 != (rc = dbenv->open(dbenv, dbedir, DB_INIT_MPOOL |
			DB_THREAD | DB_CREATE | DB_INIT_CDB, 0)))
		diex("fail: dbenv open: %s, %s",
				db_strerror(rc), dbedir);

	/* Now the database itself. */

	if (0 != db_create(&db_state.dbp, dbenv, 0))
		diex("fail: dbcreate");
	dbp = db_state.dbp;

	/* Finally, open the underlying database. */

	dbflags = DB_CREATE | DB_THREAD;
	if (flags & NLPCDB_RDONLY) {
		dbflags = DB_RDONLY;
		dddbg("%s: read-only database", __func__);
	} else 
		dddbg("%s: read-write database", __func__);

	for (;;) {
		rc = dbp->open(dbp, NULL, dbfile, NULL, DB_BTREE, 
				dbflags, 0644);
		if (0 == rc)
			break;
		else if (DB_LOCK_NOTGRANTED == rc)
			continue;
		else if (DB_LOCK_DEADLOCK == rc)
			continue;
		diex("fail: database open: %s, %s",
				db_strerror(rc), dbfile);
	}

	dddbg("%s: initialised", __func__);
}


void
db_teardown(void)
{
	int 		rc;

	if (NULL != db_state.dbp && 0 != 
			(rc = db_state.dbp->close(db_state.dbp, 0)))
		diex("%s: fail: dbclose: %s, %s", __func__,
				db_strerror(rc), db_state.dbfile);
	if (NULL != db_state.dbenv && 0 != 
			(rc = db_state.dbenv->close(db_state.dbenv, 0)))
		diex("%s: fail: dbenvclose: %s, %s", __func__,
				db_strerror(rc), db_state.dbenvdir);

	db_state.dbp = NULL;
	db_state.dbenv = NULL;
	memset(db_state.dbfile, 0, MAXPATHLEN);
	memset(db_state.dbenvdir, 0, MAXPATHLEN);

	dddbg("%s: torn down", __func__);
}


int
db_del(const struct nlpcdbrp *pair)
{
	int 		rc;
	DBT 		key;
	struct nlpcdbrp	raw;

	assert(NULL != pair);

	memcpy(raw.hash, pair->hash, NLPCR_HASHLEN);
	memset(&key, 0, sizeof(DBT));

	key.data = raw.hash;
	key.size = NLPCR_HASHLEN;

	for (;;) {
		assert(NULL != db_state.dbp);
		rc = db_state.dbp->del(db_state.dbp, NULL, 
				&key, 0);

		if (0 == rc)
			return(1);
		else if (DB_LOCK_NOTGRANTED == rc)
			continue;
		else if (DB_LOCK_DEADLOCK == rc)
			continue;
		else if (DB_NOTFOUND == rc)
			return(0);
		break;
	}

	diex("%s: fail: database delete: %s, %s", __func__,
			db_strerror(rc), db_state.dbfile);
	/*NOTREACHED*/
}


int
db_put(struct nlpcdbrp *pair)
{
	int 		rc;
	DBT 		key, data;
	struct nlpcdbr	raw;

	assert(NULL != pair);

	memset(&key, 0, sizeof(DBT));
	memset(&data, 0, sizeof(DBT));

	copyin(&pair->record, &raw);

	key.data = pair->hash;
	key.size = NLPCR_HASHLEN;

	data.data = &raw;
	data.size = sizeof(struct nlpcdbr);

	for (;;) {
		assert(NULL != db_state.dbp);
		rc = db_state.dbp->put(db_state.dbp, NULL, 
				&key, &data, 0);
		if (0 == rc)
			return(1);
		else if (DB_LOCK_NOTGRANTED == rc)
			continue;
		else if (DB_LOCK_DEADLOCK == rc)
			continue;
		else if (DB_KEYEXIST == rc)
			return(0);
		break;
	}

	diex("%s: fail: database modify: %s, %s", __func__,
			db_strerror(rc), db_state.dbfile);
	/* NOTREACHED */
}


int
db_add(struct nlpcdbrp *pair)
{
	int 		rc;
	DBT 		key, data;
	struct nlpcdbr	raw;

	assert(NULL != pair);

	memset(&key, 0, sizeof(DBT));
	memset(&data, 0, sizeof(DBT));

	copyin(&pair->record, &raw);

	key.data = pair->hash;
	key.size = NLPCR_HASHLEN;

	data.data = &raw;
	data.size = sizeof(struct nlpcdbr);

	for (;;) {
		assert(NULL != db_state.dbp);
		rc = db_state.dbp->put(db_state.dbp, NULL, 
				&key, &data, DB_NOOVERWRITE);
		if (0 == rc)
			return(1);
		else if (DB_LOCK_NOTGRANTED == rc)
			continue;
		else if (DB_LOCK_DEADLOCK == rc)
			continue;
		else if (DB_KEYEXIST == rc)
			return(0);
		break;
	}

	diex("%s: fail: database modify: %s, %s", __func__,
			db_strerror(rc), db_state.dbfile);
	/* NOTREACHED */
}


void
db_query_begin(void **cookie)
{
	DBC		*dbc;
	int		rc;

	assert(NULL != cookie);
	*cookie = NULL;

	for (;;) {
		assert(NULL != db_state.dbp);
		rc = db_state.dbp->cursor(db_state.dbp, 
				NULL, &dbc, 0);
		if (0 == rc) {
			*cookie = (void *)dbc;
			return;
		} else if (DB_LOCK_NOTGRANTED == rc)
			continue;
		else if (DB_LOCK_DEADLOCK == rc)
			continue;
		break;
	}

	diex("%s: fail: cursor: %s, %s", __func__,
			db_strerror(rc), db_state.dbfile);
	/*NOTREACHED*/
}


/*
 * Query that the state has bit-wise matches in the record.
 */
static int
query_state(const struct nlpcdbr *raw, uint64_t state)
{
	uint64_t	s, i;

	assert(NULL != raw);

	s = _ntohll(raw->state);
	for (i = 0; i < 64; i++)
		if ((state & (1 << i)) && (s & (1 << i)))
			return(1);
	return(0);
}


static int
query_auth(const struct nlpcdbr *raw, struct uri *addr)
{
	assert(NULL != raw);
	assert(NULL != raw);
	return(0 == strncmp(raw->uri.auth, addr->auth, 
				NLPCR_URI_AUTHSZ));
}


int
db_query_lookup(void **cookie, struct db_query *query,
		struct nlpcdbrp *pair)
{
	struct nlpcdbr	*raw;
	int		rc;
	DBC		*dbc;
	DBT		key, data;

	assert(NULL != cookie);
	assert(NULL != *cookie);
	assert(NULL != pair);
	assert(NULL != query);

	if (0 == query->window)
		return(-1);

	memset(&key, 0, sizeof(DBT));
	memset(&data, 0, sizeof(DBT));

	key.data = pair->hash;
	key.size = NLPCR_HASHLEN;
	dbc = *cookie;

	for (;;) {
		rc = dbc->c_get(dbc, &key, &data, DB_SET);
		if (DB_LOCK_NOTGRANTED == rc)
			continue;
		else if (DB_LOCK_DEADLOCK == rc)
			continue;
		else if (DB_NOTFOUND == rc)
			return(0);
		else if (DB_NO_ERROR == rc)
			break;
		diex("%s: fail: lookup: %s, %s", __func__,
				db_strerror(rc), db_state.dbfile);
	}

	--(query->window);
	raw = (struct nlpcdbr *)data.data;
	if (query->field & DB_FIELD_STATE) 
		if ( ! query_state(raw, query->state))
			return(0);
	if (query->field & DB_FIELD_AUTH)
		if ( ! query_auth(raw, &query->addr))
			return(0);

	memset(&pair->record, 0, sizeof(struct nlpcdbr));
	copyout(raw, &pair->record);

	return(1);
}


int
db_query_next(void **cookie, struct db_query *q,
		struct nlpcdbrp *pair)
{
	struct nlpcdbr	*raw;
	int		rc;
	DBC		*dbc;
	DBT		key, data;

	assert(NULL != cookie);
	assert(NULL != *cookie);
	assert(NULL != pair);
	assert(NULL != q);

	memset(&key, 0, sizeof(DBT));
	memset(&data, 0, sizeof(DBT));

	dbc = *(DBC **)cookie;

	for (;;) {
		if (0 == q->window) {
			dddbg("%s: window overrun", __func__);
			return(-1);
		} else do {
			rc = dbc->c_get(dbc, &key, &data, DB_NEXT);
			if (DB_LOCK_DEADLOCK == rc)
				continue;
			else if (DB_LOCK_NOTGRANTED == rc)
				continue;
			else if (DB_NOTFOUND == rc)
				return(0);
			else if (DB_NO_ERROR != rc)
				break;
			++(q->pos);
			q->skip -= (0 == q->skip ? 0 : 1);
		} while (0 != q->skip);

		if (DB_NO_ERROR != rc)
			break;

		assert(data.size == sizeof(struct nlpcdbr));

		q->window -= (-1 == q->window ? 0 : 1);
		raw = (struct nlpcdbr *)data.data;

		if (q->field & DB_FIELD_STATE) 
			if ( ! query_state(raw, q->state))
				continue;
		if (q->field & DB_FIELD_AUTH)
			if ( ! query_auth(raw, &q->addr))
				continue;

		memset(pair, 0, sizeof(struct nlpcdbrp));
		copyout(raw, &pair->record);
		memcpy(pair->hash, (unsigned char *)key.data,
				NLPCR_HASHLEN);
		return(1);

	}

	diex("%s: fail: sequence: %s, %s", __func__,
			db_strerror(rc), db_state.dbfile);
	/*NOTREACHED*/
}


void
db_query_close(void **cookie)
{
	int		rc;
	DBC		*dbc;

	assert(NULL != cookie);
	assert(NULL != *cookie);

	dbc = *(DBC **)cookie;

	for (;;) {
		if (0 == (rc = dbc->c_close(dbc))) {
			*cookie = NULL;
			return;
		} else if (DB_LOCK_NOTGRANTED == rc)
			continue;
		else if (DB_LOCK_DEADLOCK == rc)
			continue;
		break;
	}

	diex("%s: fail: cursor close: %s, %s", __func__,
			db_strerror(rc), db_state.dbfile);
	/*NOTREACHED*/
}


void
db_query_commit(void **cookie, struct nlpcdbrp *pair)
{
	struct nlpcdbr	raw;
	int		rc;
	DBC		*dbc;
	DBT		key, data;

	assert(NULL != cookie);

	memset(&key, 0, sizeof(DBT));
	memset(&data, 0, sizeof(DBT));

	key.data = pair->hash;
	key.size = NLPCR_HASHLEN;

	copyin(&pair->record, &raw);
	data.data = &raw;
	data.size = sizeof(struct nlpcdbr);

	for (;;) {
		if (NULL == (dbc = *(DBC **)cookie)) {
			dddbg("warn: %s: cookie unset", __func__);
			return;
		}
		rc = dbc->c_put(dbc, &key, &data, DB_CURRENT);
		if (0 == rc)
			return;
		else if (DB_LOCK_NOTGRANTED == rc)
			continue;
		else if (DB_LOCK_DEADLOCK == rc)
			continue;
		break;
	}

	diex("%s: fail: cursor put: %s, %s", __func__,
			db_strerror(rc), db_state.dbfile);
	/*NOTREACHED*/
}
