/* $Id: nlpcd.c,v 1.4 2007-06-06 14:07:22 kristaps Exp $ */
/*
Copyright (c) 2007 Kristaps Dzonsons
Latvijas Universitaates Matemaatikas un Informaatikas Instituuts
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

* Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
* Neither the name of the copyright holders nor the names of
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/time.h>
#include <sys/types.h>

#include <assert.h>
#include <err.h>
#include <getopt.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>


struct arg_t 
{
	int		dbg;
	int		monitor;
	char		*cache;
	char		*acache;
	char		*rep;
	char		*db;

	char		*auth;
	char		*scheme;
	char		*charset;
	char		*lang;
	char		*dict;
	
	char		**argv;
	size_t		argc;
};

#define PROC_UP		"../nlpc-up"
#define PROC_VRFY	"../nlpc-vrfy"
#define PROC_SCAN	"../nlpc-scan"
#define PROC_FETCH	"../nlpc-fetch"
#define PROC_RESOLV	"../nlpc-resolv"

#define PID_SCAN	(0)
#define PID_RESOLV	(1)
#define PID_FETCH	(2)
#define PID_VRFY	(3)
#define PID_UP		(4)
#define PID__MAX	(5)

static pid_t		children[PID__MAX];



void
dosig(int arg)
{
	(void)arg;
}


static int
startup_up(const struct arg_t *args)
{
	pid_t		pid;
	char		*argv[256];
	size_t		i, dbg;

	assert(NULL != args);
	i = 0;

	argv[i++] = PROC_UP;
	dbg = args->dbg;
	while (0 < dbg--)
		argv[i++] = "-d";

	if (NULL != args->cache) {
		argv[i++] = "-f";
		argv[i++] = args->cache;
	}

	if (NULL != args->acache) {
		argv[i++] = "-a";
		argv[i++] = args->acache;
	}

	if (NULL != args->db) {
		argv[i++] = "-p";
		argv[i++] = args->db;
	}

	argv[i] = 0;

	if (-1 == (pid = fork())) 
		return(0);
	if (0 != (children[PID_UP] = pid))
		return(1);

	execvp(argv[0], argv);

	err(EXIT_FAILURE, "%s: fail: %s", __func__, argv[0]);
	/* NOTREACHED */
}


static int
startup_vrfy(const struct arg_t *args)
{
	pid_t		pid;
	char		*argv[256];
	size_t		i, dbg;

	assert(NULL != args);
	i = 0;

	argv[i++] = PROC_VRFY;
	dbg = args->dbg;
	while (0 < dbg--)
		argv[i++] = "-d";

	if (NULL != args->charset) {
		argv[i++] = "--filter-charset";
		argv[i++] = args->charset;
	}

	if (NULL != args->dict) {
		argv[i++] = "--filter-dict";
		argv[i++] = args->dict;
	}

	if (NULL != args->cache) {
		argv[i++] = "-f";
		argv[i++] = args->cache;
	}

	if (NULL != args->db) {
		argv[i++] = "-p";
		argv[i++] = args->db;
	}

	argv[i] = 0;

	if (-1 == (pid = fork())) 
		return(0);
	if (0 != (children[PID_VRFY] = pid))
		return(1);

	execvp(argv[0], argv);

	err(EXIT_FAILURE, "%s: fail: %s", __func__, argv[0]);
	/* NOTREACHED */
}


static int
startup_fetch(const struct arg_t *args)
{
	pid_t		pid;
	char		*argv[256];
	size_t		i, dbg;

	assert(NULL != args);
	i = 0;

	argv[i++] = PROC_FETCH;
	dbg = args->dbg;
	while (0 < dbg--)
		argv[i++] = "-d";

	if (args->monitor)
		argv[i++] = "-m";

	if (NULL != args->charset) {
		argv[i++] = "--head-charset";
		argv[i++] = args->charset;
		argv[i++] = "--filter-charset";
		argv[i++] = args->charset;
	}

	if (NULL != args->lang) {
		argv[i++] = "--head-lang";
		argv[i++] = args->lang;
		argv[i++] = "--filter-lang";
		argv[i++] = args->lang;
	}

	if (NULL != args->auth) {
		argv[i++] = "--filter-auth";
		argv[i++] = args->auth;
	}

	if (NULL != args->scheme) {
		argv[i++] = "--filter-scheme";
		argv[i++] = args->scheme;
	}

	if (NULL != args->cache) {
		argv[i++] = "-f";
		argv[i++] = args->cache;
	}

	if (NULL != args->rep) {
		argv[i++] = "-r";
		argv[i++] = args->rep;
	}

	if (NULL != args->db) {
		argv[i++] = "-p";
		argv[i++] = args->db;
	}

	argv[i] = 0;

	if (-1 == (pid = fork())) 
		return(0);
	if (0 != (children[PID_FETCH] = pid))
		return(1);

	execvp(argv[0], argv);

	err(EXIT_FAILURE, "%s: fail: %s", __func__, argv[0]);
	/* NOTREACHED */
}


static int
startup_resolv(const struct arg_t *args)
{
	pid_t		pid;
	char		*argv[256];
	size_t		i, dbg;

	assert(NULL != args);
	i = 0;

	argv[i++] = PROC_RESOLV;
	dbg = args->dbg;
	while (0 < dbg--)
		argv[i++] = "-d";

	if (NULL != args->db) {
		argv[i++] = "-p";
		argv[i++] = args->db;
	}

	argv[i] = 0;

	if (-1 == (pid = fork())) 
		return(0);
	if (0 != (children[PID_RESOLV] = pid))
		return(1);

	execvp(argv[0], argv);

	err(EXIT_FAILURE, "%s: fail: %s", __func__, argv[0]);
	/* NOTREACHED */
}


static int
startup_scan(const struct arg_t *args)
{
	pid_t		pid;
	char		**argv;
	size_t		i, j, dbg;

	assert(NULL != args);
	argv = calloc(args->argc + 256, sizeof(char *));
	if (NULL == argv) {
		warn("%s: fail: allocate", __func__);
		return(0);
	}

	i = 0;

	argv[i++] = PROC_SCAN;
	dbg = args->dbg;
	while (0 < dbg--)
		argv[i++] = "-d";

	if (NULL != args->auth) {
		argv[i++] = "--filter-auth";
		argv[i++] = args->auth;
	}

	if (NULL != args->scheme) {
		argv[i++] = "--filter-scheme";
		argv[i++] = args->scheme;
	}

	if (NULL != args->db) {
		argv[i++] = "-p";
		argv[i++] = args->db;
	}

	if (NULL != args->cache) {
		argv[i++] = "-f";
		argv[i++] = args->db;
	}

	for (j = 0; j < args->argc; j++)
		argv[i++] = args->argv[j];

	argv[i] = 0;

	if (-1 == (pid = fork())) {
		free(argv);
		return(0);
	}
	if (0 != (children[PID_SCAN] = pid)) {
		free(argv);
		return(1);
	}

	execvp(argv[0], argv);

	err(EXIT_FAILURE, "%s: fail: %s", __func__, argv[0]);
	/* NOTREACHED */
}


static int
teardown(int rc)
{
	int		i;

	if (SIG_ERR == signal(SIGCHLD, SIG_IGN)) {
		warn("%s: fail: signal", __func__);
		return(EXIT_FAILURE);
	} else for (i = 0; i < PID__MAX; i++)
		if (-1 != children[i])
			kill(children[i], SIGTERM);

	return(rc);
}


static int
signal_postwait(int sig, int *rc)
{
	switch (sig) {
	case (SIGALRM):
		return(1);
	case (SIGTERM):
		/* FALLTHROUGH */
	case (SIGINT):
		*rc = EXIT_SUCCESS;
		break;
	case (SIGCHLD):
		*rc = EXIT_FAILURE;
		break;
	}
	return(0);
}


static int
signal_wait(sigset_t *p, long tm, int *rc)
{
	struct itimerval itm;
	int		sig;

	assert(NULL != p);
	assert(NULL != rc);

	itm.it_interval.tv_sec = 0;
	itm.it_interval.tv_usec = 0;
	itm.it_value.tv_sec = tm;
	itm.it_value.tv_usec = 0;

	if (0 != tm && -1 == setitimer(ITIMER_REAL, &itm, NULL)) {
		warn("%s: fail: setitimer", __func__);
		return(0);
	}

	if (0 != sigwait(p, &sig))
		warn("%s: fail: sigwait", __func__);
	else
		return(signal_postwait(sig, rc));

	return(0);
}


static int
signal_block(sigset_t *set)
{
	assert(NULL != set);

	if (SIG_ERR == signal(SIGCHLD, dosig))
		warn("%s: fail: signal", __func__);
	else if (-1 == sigemptyset(set))
		warn("%s: fail: sigemptyset", __func__);
	else if (-1 == sigaddset(set, SIGCHLD))
		warn("%s: fail: sigaddset", __func__);
	else if (-1 == sigaddset(set, SIGINT))
		warn("%s: fail: sigaddset", __func__);
	else if (-1 == sigaddset(set, SIGTERM))
		warn("%s: fail: sigaddset", __func__);
	else if (-1 == sigaddset(set, SIGALRM))
		warn("%s: fail: sigaddset", __func__);
	else if (-1 == sigprocmask(SIG_BLOCK, set, NULL))
		warn("%s: fail: sigprocmask", __func__);
	else 
		return(1);
	return(0);
}


int
main(int argc, char *argv[])
{
	int 		c, rc;
	struct arg_t	args;

	extern char	*__progname, *optarg;
	extern int	optind;
	sigset_t	sset;

	static struct option longopts[] = {
		{ "filter-charset", 	required_argument, NULL, 1 },
		{ "filter-lang", 	required_argument, NULL, 2 },
		{ "filter-scheme", 	required_argument, NULL, 3 },
		{ "filter-auth", 	required_argument, NULL, 4 },
		{ "filter-dict", 	required_argument, NULL, 5 },
		{ NULL, 0, NULL, 0 }
	};

	memset(children, -1, sizeof(children));
	memset(&args, 0, sizeof(struct arg_t));

	while (-1 != (c = getopt_long(argc, argv, "a:f:dhmp:r:v", 
					longopts, NULL))) 
		switch (c) {
		case (1):
			args.charset = optarg;
			break;
		case (2):
			args.lang = optarg;
			break;
		case (3):
			args.scheme = optarg;
			break;
		case (4):
			args.auth = optarg;
			break;
		case (5):
			args.dict = optarg;
			break;
		case ('a'):
			args.acache = optarg;
			break;
		case ('d'):
			if (args.dbg < 3)
				(args.dbg)++;
			break;
		case ('f'):
			args.cache = optarg;
			break;
		case ('m'):
			args.monitor = 1;
			args.dbg = 0;
			break;
		case ('p'):
			args.db = optarg;
			break;
		case ('r'):
			args.rep = optarg;
			break;
		case ('v'):
			fprintf(stdout, "%s version %s\n", 
					__progname, VERSION);
			return(EXIT_SUCCESS);
		case ('h'):
			/*FALLTHROUGH*/
		default:
			fprintf(stderr, "usage: %s [-dhmv] [-a path] "
					"[-f path] [-p path] "
					"[LONG_OPTS]\n", __progname);
			return(EXIT_FAILURE);
		}
	argc -= optind;
	argv += optind;

	args.argc = argc;
	args.argv = argv;

	rc = EXIT_FAILURE;

	if ( ! signal_block(&sset))
		goto main_tail;

	if ( ! startup_scan(&args))
		goto main_tail;
	if ( ! signal_wait(&sset, 1, &rc))
		goto main_tail;

	if ( ! startup_resolv(&args))
		goto main_tail;
	if ( ! signal_wait(&sset, 1, &rc))
		goto main_tail;

	if ( ! startup_fetch(&args))
		goto main_tail;
	if ( ! signal_wait(&sset, 1, &rc))
		goto main_tail;
	
	if ( ! startup_vrfy(&args))
		goto main_tail;
	if ( ! signal_wait(&sset, 1, &rc))
		goto main_tail;

	if ( ! startup_up(&args))
		goto main_tail;
	if ( ! signal_wait(&sset, 1, &rc))
		goto main_tail;

	rc = EXIT_SUCCESS;

	signal_wait(&sset, 0, &rc);

main_tail:

	return(teardown(rc));
}

