	/* $Id: nlpcrawl.h,v 1.108 2007-06-05 23:14:44 kristaps Exp $ */
/*
Copyright (c) 2007 Kristaps Dzonsons
Latvijas Universitaates Matemaatikas un Informaatikas Instituuts
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

* Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
* Neither the name of the copyright holders nor the names of
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef NLPCRAWL_H
#define NLPCRAWL_H

__BEGIN_DECLS


/* 
 * Signal for enlarging a scan window. 
 */
#define NLPC_SIGNAL_EUP		(SIGUSR1)

/* 
 * Signal for narrowing a scan window.
 */
#define NLPC_SIGNAL_EDOWN	(SIGUSR2)

/* 
 * Signal for triggering scans.
 */
#define NLPC_SIGNAL_SCAN	(SIGHUP)

/* 
 * Signal for triggering "safe shutdown".
 */
#define NLPC_SIGNAL_TERM	(SIGTERM)

/* 
 * Default location of database files. 
 */
#define NLPC_DEFAULT_DIR	"/tmp/nlpcrawl"

#define NLPC_DEFAULT_REPS	"/tmp/nlpcrawl/cache/reps"

/* 
 * Default location of downloaded cache. 
 */
#define NLPC_DEFAULT_FILES	"/tmp/nlpcrawl/cache"

/*
 * Default location of file archive.
 */
#define NLPC_DEFAULT_ARCHIVE	"/tmp/nlpcrawl/archive"

/* 
 * Length of a URI hash.  These are represented by MD5 digests, which
 * are 16 bytes in width.
 */
#define NLPCR_HASHLEN		(16)

/* 
 * Maximum length of the cache directory prefix (less constructed path
 * from hash).
 */
#define NLPCR_CACHEDIRLEN	(MAXPATHLEN - (NLPCR_HASHLEN * 3))

/* 
 * Maximum length of a language charset.  This is from eyeballing the
 * output of iconv -l.
 */
#define NLPCR_CHARSETLEN	(32)

/* 
 * See RFC 1766, 2616.  Note that this allows multiple languages per
 * page; all tokens are stored here.  This may be overrun in the event
 * of too many languages. 
 */
#define NLPCR_LANGLEN		(32)

/* 
 * The following deal with URI field lenghts.  None of these should
 * actually be fixed; be conservative in sizing.
 */
#define NLPCR_URI_SCHEMESZ	(16)
#define NLPCR_URI_AUTHSZ	(256)
#define NLPCR_URI_QUERYSZ	(1024)
#define NLPCR_URI_PATHSZ	(1024)
#define NLPCR_URI_FRAGSZ	(128)

/*
 * Sum of all URI parts plus miscellaneous delimiters.
 */
#define NLPCR_URI_MAXSZ		(NLPCR_URI_SCHEMESZ + \
				 NLPCR_URI_AUTHSZ + \
				 NLPCR_URI_QUERYSZ + \
				 NLPCR_URI_PATHSZ + \
				 NLPCR_URI_FRAGSZ + 32)

#define NLPCR_URI_FMTSZ		(NLPCR_URI_MAXSZ + 10)

#define NLPCR_IP_WIDTH		(4)
#define NLPCR_IP_STRING		(16)


/* 
 * A URI conforms to RFC 2986/Std. 66.  See the RFC for relevant
 * terminology.
 */
struct uri {
	char		scheme[NLPCR_URI_SCHEMESZ];
	char		auth[NLPCR_URI_AUTHSZ];
	char		path[NLPCR_URI_PATHSZ];
	char		query[NLPCR_URI_QUERYSZ];
	char		fragment[NLPCR_URI_FRAGSZ];
};


/* 
 * This is documented in full in nlpcrawl(1).  Consider the manual to be
 * canonical.
 */
struct nlpcdbr {
	struct uri 	uri;
	/*
	 * States are described in nlpcrawl(1), although the names are a
	 * little bit different (NLPCDBR_FRESH <-> FRESH, etc.).
	 */
#define NLPCDBR_NEW	(1 << 0)
#define NLPCDBR_FRESH	(1 << 1)
#define NLPCDBR_DIRTY	(1 << 2)
#define NLPCDBR_UNREACH	(1 << 3)
#define NLPCDBR_DEAD	(1 << 4)
#define NLPCDBR_SCAN	(1 << 5)
#define NLPCDBR_READY	(1 << 6)
	/* (called "RESOLV" in nlpcrawl.1) */
#define NLPCDBR_VRFY	(1 << 7)
	uint64_t 	state;

#define NLPCDBR_OK	(0x0000)	/* All's well.               */
#define NLPCDBR_GENERAL	(0x0001)	/* A general/unnamed error.  */
#define NLPCDBR_RESOLV	(0x0002)	/* Name resolution error.    */
#define NLPCDBR_CONNECT	(0x0003)	/* Connect error.            */
#define NLPCDBR_TIMEO	(0x0004)	/* Socket timeout.           */
#define NLPCDBR_SERVER	(0x0005)	/* Server-side error.        */
#define NLPCDBR_SYNTAX	(0x0006)	/* Data syntax malformed.    */
#define NLPCDBR_URL	(0x0101)	/* URI syntax malformed.     */
#define NLPCDBR_MOVED	(0x0102)	/* Data moved.               */
#define NLPCDBR_CTYPE	(0x0103)	/* Content-media invalid.    */
#define NLPCDBR_CHARSET	(0x0201)	/* Content-charset invalid.  */
#define NLPCDBR_LANG	(0x0202)	/* Language invalid.         */
#define NLPCDBR_RTD	(0x0301)	/* Revivals exceeded.        */
#define NLPCDBR_LIMIT	(0x0400)	/* Download-limit exceeded.  */
#define NLPCDBR_REP	(0x0401)	/* Denied as per REP.        */
	uint64_t 	code;

	uint64_t	revived;
	
	unsigned char	ip[NLPCR_IP_WIDTH];

	uint64_t 	ctime;
	uint64_t 	mtime;
	/* TODO: remove this: unnecessary. */
	uint64_t 	atime;
	uint64_t	ftime;

	unsigned char	data[NLPCR_HASHLEN];
	uint64_t	size;

#define NLPCDBR_TYPE_NONE (1 << 0)
#define NLPCDBR_TYPE_TXT_XML (1 << 1)	/* text/xml */
#define NLPCDBR_TYPE_TXT_HTML (1 << 2)	/* text/html */
#define NLPCDBR_TYPE_APP_XML (1 << 3)	/* application/xml */
#define NLPCDBR_TYPE_APP_XHTML (1 << 4)	/* application/xml+xhtml */
#define NLPCDBR_TYPE_TXT_PLAIN (1 << 5)	/* text/plain */
	uint64_t 	type;

	char 		charset[NLPCR_CHARSETLEN];
	char		lang[NLPCR_LANGLEN];
	uint64_t 	expires;
};


/* 
 * Returns static buffer with nil-terminated string representation of a
 * record state.
 */
extern char *
nlpcdbr_state2str(uint64_t state);


/*
 * Like nlpcdbr_state2str but thread-safe.  Buffer should be greater
 * than 32 bytes wide.
 */
extern char *
nlpcdbr_state2str_r(char *str, uint64_t state);


/*
 * Convert a string representation of a state (probably as made by
 * nlpcdbr_state2str) back into integral form.
 */
extern uint64_t
nlpcdbr_str2state(const char *state);


/* 
 * Returns static buffer with nil-terminated string representation of a
 * record content/media type.
 */
extern char *
nlpcdbr_type2str(uint64_t type);


/*
 * Like nlpcdbr_type2str but thread-safe.  Buffer should be greater than
 * 32 bytes wide.
 */
extern char *
nlpcdbr_type2str_r(char *str, uint64_t type);


/* 
 * Returns static buffer with nil-terminated string representation of a
 * record error code.
 */
extern char *
nlpcdbr_code2str(uint64_t code);


/*
 * Like nlpcdbr_code2str but thread-safe.  Buffer should be greater than
 * 32 bytes wide.
 */
extern char *
nlpcdbr_code2str_r(char *str, uint64_t code);


/* 
 * Returns static buffer with nil-terminated string representation of a
 * digest field.
 */
extern char *
nlpcdbr_hash2str(const unsigned char hash[NLPCR_HASHLEN]);


/*
 * Like nlpcdbr_hash2str but thread-safe.  Buffer should be greater than
 * (NLPCR_HASHLEN * 2) + 1 bytes wide.
 */
extern char *
nlpcdbr_hash2str_r(char *str, const unsigned char hash[NLPCR_HASHLEN]);


/*
 * Convert a hash into a filename structure under prefix.  The prefix
 * must be of length such that it and the hash suffix may fit into a
 * MAXPATHLEN.  The returned buffer is a static buffer that will be
 * overwritten on next usage.  Its length is MAXPATHLEN and is
 * guaranteed to be nil-terminated.
 */
extern char *
nlpcdbr_hash2path(const char *prefix,
		const unsigned char hash[NLPCR_HASHLEN]);


/*
 * Like nlpcdbr_hash2path but reentrant.  
 */
extern char *
nlpcdbr_hash2path_r(char output[MAXPATHLEN], const char *prefix,
		const unsigned char hash[NLPCR_HASHLEN]);


/*
 * Convert a stringified hash (probably as made by nlpcdbr_hash2str)
 * back into a non-nil-terminated hash of length NLPCR_HASHLEN.
 */
extern char *
nlpcdbr_str2hash(const char *str);


/*
 * Like nlpcdbr_str2hash but thread-safe.  Buffer should be greater than
 * NLPCR_HASHLEN bytes wide.
 */
extern char *
nlpcdbr_str2hash_r(char *buf, const char *str);



/*
 * A convenient hash/record pair structure.  These absolutely must be
 * coherent with one another (note there are times when one part of the
 * structure will be valid while another won't: as in db_lookup).
 */
struct nlpcdbrp {
	struct nlpcdbr	record;
	unsigned char	hash[NLPCR_HASHLEN];
};


/*
 * Zero out then construct a new record.  The record's states are all
 * state to the initial ones, and the pair's hash appropriate generated.
 */
extern void
nlpcr_new(const struct uri *addr, struct nlpcdbrp *res);


void
nlpcr_archive(const char *prefix, const char *aprefix,
		const struct nlpcdbrp *pair);



/*
 * Parse a uri, whether it be absolute or relative.  Returns 0 on
 * failure, 1 on success.  If urisz is 0, the string length of uri is
 * used instead (which may be 0).  The resulting uri will be zeroed
 * (even in the event of failure). This follows rfc 3096/std 66.  Note
 * that the parse methods are not validating, for example, a uri may
 * parse as absolute but contain invalid escape codes.  However, the
 * methods are auto-normalising, meaning the following two uri
 * structures are equivalent:
 *
 *     hTTp://g%4Fogle.COM/../..///index.html
 *     http://google.com/index.html
 */
extern int
uri_parse(const char *uri, size_t urisz, struct uri *p);


/*
 * Decode superfluous percent-encodings within a string.  Returns 0 if
 * the string is not appropriately percent-encoded.
 */
extern int
uri_pct_decode(char *p, size_t sz);


/*
 * Return 1 if the uri is an absolute-uri (as defined by rfc 3096/std 66
 * abnf).  This defines a subset of uri without ending fragment and with
 * a scheme.
 */
extern int
uri_is_abs(const struct uri *p);


/*
 * Resolve a relative uri referenced based upon an absolute baseuri.
 * The resulting absolute uri will be placed in p.  Note that the result
 * may be a relative uri (if, e.g., a fragment is specified).  This
 * follows the uri resolution algorithm laid forth in rfc 3096 sec.
 * 5.2.2.
 */
extern int
uri_ref_resolve(const struct uri *base, const struct uri *rel,
		struct uri *p);


/*
 * "Ascend" the uri.  When the uri mayn't be ascended any more, returns
 * zero.
 */
extern int
uri_ascend(struct uri *p);


/*
 * Format a uri to be pretty (or in other words, the components are
 * easily discernable).  This returns a static buffer that remains
 * consistent until the next time uri_topstring is called.  This
 * function shouldn't be used in any threads.
 * function is not thread-safe.
 */
extern char *
uri_topstring(const struct uri *p);


/*
 * Exactly as uri_topstring but thread-safe.  Returns the same pointer
 * as buf (in order to behave as uri_topstring).  The buf parameter
 * should be at least of length NLPCR_URI_FMTSZ.
 */
extern char *
uri_topstring_r(const struct uri *p, char *buf);


/*
 * Format a uri.  This formats the uri in canonical mode.  This returns
 * a static buffer.
 */
extern char *
uri_tostring(const struct uri *p);


extern char *
uri_ipmerge_r(const struct uri *p, 
		const unsigned char *ip, char *buf);


/*
 * Format an IP address.  Buffer should be greater than NLPCR_IP_STRING
 * bytes wide.
 */
extern char *
ip_tostring_r(const unsigned char *ip, char *s);


/*
 * Like uri_tostring but thread-safe.  Buffer should be greater than
 * NLPCR_URI_MAXSZ bytes wide.
 */
extern char *
uri_tostring_r(char *str, const struct uri *p);


/*-------------------------------------------------------------------*/


/*
 * The integral debug level.  1 lets through dbg; 2, also ddbg; 3, also
 * dddbg.  The default value is 0, which displays no debugging data.
 * This should NOT be manipulated when multiple threads are in use; it's
 * not thread-safe.
 */
extern int
dbg_lvl;


/*
 * Informational messages.  See vdbg.
 */
extern void
dbg(const char *fmt, ...) __attribute__((format (printf, 1, 2)));


/*
 * "Debugging" messages (not significant to the operator).  See vdbg.
 */
extern void
ddbg(const char *fmt, ...) __attribute__((format (printf, 1, 2)));


/*
 * Verbose-debugging messages.  See vdbg.
 */
extern void
dddbg(const char *fmt, ...) __attribute__((format (printf, 1, 2)));


/*
 * Underlying routine for all debugging functions.  This function is
 * thread-safe: the printing area is protected by a mutex.
 */
extern void
vdbg(int level, const char *fmt, va_list ap);


/*
 * Macros for abnormal exits.  These should ONLY ONLY ONLY be called in
 * case of real error (that's not abort()-worthy) as underlying
 * structures (like the database) may be left in an inconsistent state.
 */
#define diex(fmt, args...) errx (EXIT_FAILURE, fmt , ##args)
#define die(fmt, args...) err (EXIT_FAILURE, fmt , ##args)


/*-------------------------------------------------------------------*/


/*
 * Compute hash value based on a string.  The string must be
 * nil-terminated.   This is equivalent to using the md5_hash_open,
 * md5_hash_update, and md5_hash_close routines.
 *
 * This uses Ron Rivest's md5 algorithm as implemented by Colin Plumb.
 * See md5.c copyright for details.
 */
extern void
md5_hash(const char *p, unsigned char out[NLPCR_HASHLEN]);


/*
 * Open a hashing session.  See md5_hash.
 */
extern void
md5_hash_open(void **cookie);


/*
 * Close/finalise a hashing session.  Invalidates the cookie pointer.
 * See md5_hash.
 */
extern void
md5_hash_close(void **cookie, unsigned char hash[NLPCR_HASHLEN]);


/*
 * Update the hash calculated during a session.  This doesn't touch the
 * session if the input is zero-length.  See md5_hash.
 */
extern void
md5_hash_update(const char *p, size_t sz, void *cookie);


/*-------------------------------------------------------------------*/


/*
 * Construct a dictionary-hashtable (no value for keys).  These routines
 * are signal safe.  A dictionary's good for quick lookups of single
 * words.  See hash_new.
 */
extern void
dict_new(void **cookie);


/*
 * Like dict_new, but used for hashtables (instead of dictionaries).
 * Hashtables have both keys and values.
 */
extern void
hash_new(void **cookie);


#define HASH_FILE_WORD	(100)

/*
 * Fill a hashtable from a file of newline-separated words.  This
 * accepts words of up to HASH_FILE_WORD characters in length.
 * Duplicates are silently ignored.  Returns 1 on success, 0 on failure.
 */
extern int
dict_file(void **cookie, const char *file);


/*
 * Make and fill a hashtable with a comma-separated list of values.  See
 * dict_new.  str may be changed by the callee; it may be safely be
 * memory-managed by the caller.
 */
extern void
dict_fill(void **cookie, char *str);


/*
 * Count number of keys in dictionary (iterates over whole sequence).
 */
extern uint64_t
dict_keys(void **cookie);


/*
 * Query whether the hashtable has a given key.  If the hash cookie is
 * NULL, or a match is found, this returns 1, else it returns 0.
 */
extern int
dict_has(void **cookie, void *key, size_t keysz);


/*
 * Begin an iteration over a dictionary.  The seq value must be later
 * reclaimed with dict_iter_close.
 */
extern void
dict_iter_begin(void **cookie, void **seq);


/*
 * Fill in the key with the next entry, if it's found in the sequence.
 * The pointed-to location must have enough memory for the key.
 */
extern int
dict_iter_next(void **seq, void *key, size_t sz);


/*
 * Close an iteration as begun with dict_iter_begin.
 */
extern void
dict_iter_close(void **seq);


/*
 * Put a given key into the hashtable.  Zero-length keys are not
 * allowed.  Duplicate entries will be ignored.  The key will be copied
 * by the callee: the value may be memory managed by the caller.
 */
extern void
dict_put(void **cookie, void *key, size_t keysz);


/*
 * Like dict_put.  Stores the data pointer at the given location.
 * Duplicates overwrite the given location.
 */
extern void
hash_put(void **cookie, void *key, size_t keysz, void *data);


/*
 * Retrieve an element from the database.  Returns 1 if the element
 * exists, else 0 if it does not.  If data is not NULL, it's filled
 * with the found element.
 */
extern int
hash_get(void **cookie, void *key, size_t keysz, void **data);


/*
 * Delete an entry from the hashtable.  This doesn't free the pointer
 * memory, only removes its reference from the table.  Returns 1 if the
 * entry was removed, 0 if it wasn't found.
 */
extern int
hash_del(void **cookie, void *key, size_t keysz);


/*
 * Free up a hashtable created with dict_new.
 */
extern void
dict_free(void **cookie);


/*
 * Like dict_free.  Doesn't touch the actual data.
 */
extern void
hash_free(void **cookie);


/*-------------------------------------------------------------------*/


/*
 * Used when iterating over a database and extracting specific field
 * types.  One may set the window (width of scan or -1 to disable), skip
 * (offset from which to scan or 0 for current), state (bit-wise states
 * to scan for), and/or adddr (the authority to match); for the last
 * two, the field must be appropriate set.
 */
struct db_query {
	int32_t		window;		/* Scan window (-1 off).     */
	uint64_t	pos;		/* Number of items scanned.  */

#define DB_FIELD_NONE	(0)		/* No fields (all returned.  */
#define DB_FIELD_STATE	(1 << 0)	/* Match by state.           */
#define DB_FIELD_AUTH	(1 << 1)	/* Match by URI auth.        */
	uint8_t		field;		/* Which field to match.     */

	uint64_t 	state;		/* Bit-wise OR of states.    */
	struct uri 	addr;		/* Used for auth match.      */
	uint64_t	skip;		/* Skips this number of rec- */
 					/* ords (if zero, operates   */
					/* normally) (only valid for */
					/* db_query_next scans).     */
};


/*
 * Grow the scan window.  This can grow up to 63, or 2 ^ 63, which is a
 * Big Number (inclusive).
 */
extern void
db_window_grow(void);


/*
 * Shrink the scan window.  This will not shrink the window to below
 * zero (inclusive).
 */
extern void
db_window_shrink(void);


/*
 * Fill a query structure with the current scan window size.
 */
extern void
db_window_fill(struct db_query *p);


#define NLPCDB_RDONLY	(1 << 0)

/*
 * Initialise a database.  This must be called before any database
 * operations; further it /must/ be symmetric to db_teardown.  If flags
 * is se to NLPCDBR_RDONLY, the database will be read-only.
 */
extern void
db_init(const char *file, uint32_t flags);


/*
 * Safely close out a database.  This /must/ be called for every opened
 * database or corruption ensues.
 */
extern void
db_teardown(void);


/*
 * Atomically add an entry to the database.  If one exists by that
 * particular key, 0 is returned.
 */
extern int
db_add(struct nlpcdbrp *pair);


/*
 * Atomically replace an entry in the database.  If one does not exists
 * by that particular key, 0 is returned.
 */
extern int
db_put(struct nlpcdbrp *pair);


/*
 * Atomically delete an entry from the database.  If one does not exists
 * by that particular key, 0 is returned.
 */
extern int
db_del(const struct nlpcdbrp *pair);


/*
 * Begin a query sequence.
 */
extern void
db_query_begin(void **cookie);


/*
 * Like db_query_lookup, but atomically finds the matching record.
 * Returns 0 if there are no more records to return, 1 if there are, or
 * -1 if a window has been specified and exceeded.
 */  
extern int
db_query_next(void **cookie, struct db_query *query,
		struct nlpcdbrp *pair);


/*
 * Like db_query_next, but atomically finds the matching record by using
 * the hash portion of pair (the record portion is written-to).  Returns
 * 0 if there is no record to return, 1 on success, or -1 if a
 * limitation clause has been reached.
 */
extern int
db_query_lookup(void **cookie, struct db_query *query,
		struct nlpcdbrp *pair);


/*
 * Atomically commit the current query value (via db_query_lookup or
 * db_query_next).
 */
extern void
db_query_commit(void **cookie, struct nlpcdbrp *pair);


/*
 * Atomically close out a query sequence.  cookie is nullified.
 */
extern void
db_query_close(void **cookie);



/*
 * A fully-parsed http/https header.
 */
struct http_header {
	char 		*key;		/* The raw nil-term key. */
	char		*val;		/* The raw nil-term val. */

#define HTTP_UNKNOWN	(0x00)
#define HTTP_EXPIRES	(0x01)
#define HTTP_PRAGMA (0x02)
#define HTTP_CONTENT_TYPE (0x03)
#define HTTP_CONTENT_LANG (0x04)
#define HTTP_LOCATION	(0x05)
	uint64_t	type;		/* Type of known key. */

	union {
		char	*expires;	/* "Expires" date. */
		struct http_header_pragma {
#define HTTP_PRAGMA_UNKNOWN (0x00)
#define HTTP_PRAGMA_NOCACHE (0x01)
			uint64_t pragma_type; /* "Pragma". */
			union {
				int nocache; /* "Pragma-nocache". */
			} data;
		} pragma;
		struct uri location;	/* "Location" uri. */
		char 	*clang;		/* "Content-language". */
		struct http_header_content_type {
			uint64_t media;	/* "Content-type" media. */
			char *charset;	/* "Content-type" charset. */
		} ctype;
	} data;
};


/*
 * Parse a http message header line as given by an external server.
 * This is able to properly parse several specific fields as defined in
 * struct http_header.  This is not a canonical validating parser: it
 * does, however, normalise the status line in its output fields
 * (assuming success).  This conforms to rfc 2616 and the original
 * http/1.0 format.  Note that "p" is modified and may not be reliably
 * re-used.
 */
extern int
http_parse(char *p, size_t sz, struct http_header *head);


/*
 * Parse an http response status-line as given by an external server.
 * This attempts to extract the status code.  This conforms to rfc 2616
 * and the original http/1.0 format.  Note that "p" is modified and may
 * not be reliably re-used.
 */
extern int
http_status(char *p, size_t sz, uint32_t *code);


/*-------------------------------------------------------------------*/


/*
 * Like parse_hex, but parses base-10 values: see strtol.
 */
extern int
parse_int(const char *from, int *to, int max, int min);


/*
 * Like parse_int, but parses base-16 values (these may or may not have
 * the 0x prefix): see strtol.
 */
extern int
parse_hex(const char *from, int *to, int max, int min);


/*
 * Standard mkdir_p utility.  This one (c) Nate Nielsen.  See util.c for
 * details.  All segments are set to mode.  The path may be modified
 * during file creation.
 */
extern int
mkdir_p(char *path, mode_t mode);


/*
 * Get the time as a uint64_t.  Always returns appropriately.
 */
extern uint64_t
time_uint64(void);


#define time_s(p) do if ((time_t)-1 == ((p) = time(NULL)))            \
	die("%s: fail: time", __func__); while (0)

#define calloc_s(p, nmem, smem) do \
	if (NULL == ((p) = calloc((nmem), (smem))))                   \
		die("%s: fail: allocate", __func__); while (0)        \


/*-------------------------------------------------------------------*/


/*
 * Signal a process (specified by next) with a scan signal.  Errors
 * encountered when accessing the file are ignored.  Returns 1 if the
 * relevant pid was signalled, 0 otherwise.  This function is safe to be
 * used across threads, assuming that the environment (prefix, etc.)
 * remains consistent.
 */
extern int
signal_next(uint64_t state);


/*
 * If state is 0, does not install a signal file and ignores (instead of
 * blocking) relevant signals.  This should be called from one thread.
 */
extern void
signal_init(const char *prefix, uint64_t state);


/*
 * Teardown a signal environment installed by signal_init.   This should
 * be called from one thread.
 */
extern void
signal_teardown(void);


#define SIGNAL_EXIT		(1 << 0)
#define SIGNAL_SCAN		(1 << 1)
#define SIGNAL_TIMEOUT		(1 << 2)

/*
 * Wait for signals to be delivered.  If the timeout (sec) is zero, wait
 * indefinitely, else return in that time.  Returns a bit-field of
 * SIGNAL_EXIT (if an exit requested was delivered while or before
 * waiting), SIGNAL_SCAN (if a scan request was delivered while or
 * before waiting), and SIGNAL_TIMEOUT if the timeout was reached before
 * any signals were delivered.  This is safe to call across threads (but
 * you probably don't want to!).
 */
extern int
signal_wait(unsigned int sec);


/*
 * Pop any relevant signals waiting on the signal queue.  Behaves
 * similarly to signal_wait in non-blocking mode.  This is safe to call
 * across threads (but you probably don't want to!).
 */
extern int
signal_test(void);


/*-------------------------------------------------------------------*/


/*
 * Initialise a queue.  This is guaranteed to return a valid pointer.
 * Queues are LIFO.  All operations on the queue are guaranteed to
 * complete in constant time; the space required by the queue is
 * linearly proportionate to the number of entries.  No functions are
 * thread-safe.
 */
void
q_new(void **);


/*
 * Free a queue made with q_new (only the internals, not the data as
 * added with q_add).  If the queue is NULL, this merely returns.   This
 * function is not thread safe.
 */
void
q_free(void **queue);


/*
 * Append a queue node.  The data is not internally touched in any way
 * (although it may not be NULL).  This function is not thread safe.
 */
void
q_add(void **queue, void *data);


/*
 * Test whether the queue is empty.  This function is not thread safe.
 */
int
q_empty(void **queue);


/*
 * Remove a queue node from the head of the queue; return NULL if the
 * queue is empty.  This function is not thread safe.
 */
void *
q_rem(void **queue);


/*
 * Peek at the data at the head of the queue.
 */
void *
q_head(void **queue);


/*-------------------------------------------------------------------*/


/*
 * Create a synchronised queue.  This is exactly like a regular queue
 * except that operations are thread-safe.  If the indicated maximum
 * size is zero, the queue is un-bounded.  See q_new.
 */
void *
squeue_new(size_t max);


#define SQUEUE_NOTIFY		(1 << 0)
#define SQUEUE_WAIT		(1 << 1)

/*
 * Exactly like q_rem.  Blocks until a lock is acquired.  If SQUEUE_WAIT
 * is specified, this will wait for the queue to be non-empty, should it
 * be empty at the time of lock acquisition.  If SQUEUE_NOTIFY is
 * specified and the queue was full before an element was dequeued, any
 * waiting threads (via squeue_add) are notified.  Otherwise, NULL will
 * be returned.  The waiting is thread-cancellable.
 */
void *
squeue_rem(void *cookie, uint64_t flags);


/*
 * Notify any blockers waiting on the queue cookie (this can happen both
 * in squeue_rem and squeue_add).
 */
void
squeue_notify(void *cookie);


/*
 * Exactly like q_add.  Blocks until a lock is acquired.  If
 * SQUEUE_NOTIFY is specified, this will notify any waiting threads (see
 * squeue_rem) ONLY if the size bound was not exceeded.  If the size
 * bound is exceeded and SQUEUE_WAIT is specified, this will block until
 * there is room to append.  Otherwise, returns 0 if the size bound was
 * exceeded (data was not queued), 1 if the element was queued.  The
 * waiting is thread-cancellable.
 */
int
squeue_add(void *cookie, void *data, uint64_t flags);


/*
 * Exactly like q_free.  Note that this MUST not be run concurrent to
 * any other accesses (e.g., squeue_rem and squeue_add): the internal
 * functions are not protected from multi-thread access.
 */
void
squeue_free(void *cookie);


/*-------------------------------------------------------------------*/


/*
 * Note that the entries contained by allow and deny are guaranteed to
 * be a uniform width of NLPCR_URI_PATHSZ.
 */
struct rep_parse_t
{
	uint32_t	crawl_delay;	/* Crawl-delay value.        */
	void		*allow;		/* Allow queue.              */
	void		*deny;		/* Deny queue.               */
	size_t		max_allow;	/* Max size of allow queue.  */
	size_t		max_deny;	/* Max size of deny queue.   */

#define REP_DELAY	(1 << 1)
#define REP_ALLOW	(1 << 2)
#define REP_DENY	(1 << 3)
	uint64_t	fields;		/* Valid fields.             */
};


/*
 * Initialise the structures in a rep_parse_t.  Does not initialise the
 * actual pointer.  max_allow and max_deny, if non-zero, will be used to
 * limit the size of the queues in each returned structure.
 */
void
rep_parse_init(size_t max_allow, size_t max_deny, 
		struct rep_parse_t *p);


/*
 * Free the structures in a rep_parse_t.  Does not free the pointer
 * parameter.  This clears out any remaining memory allocated from the
 * parse in the queues.
 */
void
rep_parse_free(struct rep_parse_t *p);


/*
 * Parse a robots.txt (REP).  This puts the results in res.  Note that
 * the structure may be partially-filled if errors occured (0 is
 * returned), thus it's necessary to call rep_parse_free between any
 * calls to this structure.
 */
int
rep_parse(const char *host, int fd, struct rep_parse_t *res);


__END_DECLS

#endif /*!NLPCRAWL_H*/
