/* $Id: rep.c,v 1.11 2007-06-05 16:02:52 kristaps Exp $ */
/*
Copyright (c) 2007 Kristaps Dzonsons
Latvijas Universitaates Matemaatikas un Informaatikas Instituuts
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

* Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
* Neither the name of the copyright holders nor the names of
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/param.h>

#include <assert.h>
#include <err.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <inttypes.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "nlpcrawl.h"


struct rstate_t
{
	int		active;
	int		recording;

	const char	*host;
	
	const char	*buf;
	size_t		bufsz;

	size_t		sz_allow;
	size_t		sz_deny;
	
	int		comment;

	struct rep_parse_t *parser;
};


static int
parse_skiptoval(const char *key, struct rstate_t *st, 
		const char **res)
{
	char		*p;

	assert(NULL != key);
	assert(NULL != res);
	assert(NULL != st);

	if (NULL == (p = memchr(st->buf, ':', st->bufsz))) {
		dddbg("%s: fail: missing separator: %s, %s",
				__func__, key, st->host);
		return(0);
	}
	++p;
	while (9 == *p || 32 == *p)
		p++;
	if (0 == *p) {
		dddbg("%s: fail: missing value: %s, %s",
				__func__, key, st->host);
		return(0);
	}
	*res = p;
	return(1);
}


static int
parse_useragent(struct rstate_t *st)
{
	const char	*p;
	size_t		max, sz;

	assert(NULL != st);

	if ( ! parse_skiptoval("User-Agent", st, &p))
		return(0);
	max = st->bufsz - (p - st->buf);

	if (max >= (sz = strlen("nlpcrawl")) &&
			0 == strncasecmp(p, "nlpcrawl", sz))
		st->active = 1;
	else if (max >= (sz = strlen("*")) &&
			0 == strncasecmp(p, "*", sz))
		st->active = 1;
	else if (1 == st->recording)
		st->active = 0;

	st->recording = 0;
	return(1);
}


static int
parse_allow(struct rstate_t *st)
{
	const char	*p;
	char		*path, *sv;

	assert(NULL != st);
	st->recording = 1;

	if (0 == st->active)
		return(1);
	if ( ! parse_skiptoval("Allow", st, &p))
		return(0);
	assert(NULL != p);

	assert(NULL != st->parser);
	if (st->parser->max_allow < st->sz_allow + 1 &&
			0 != st->parser->max_allow) {
		dddbg("%s: fail: allow-path queue limit: %s",
				__func__, st->host);
		return(0);
	}

	calloc_s(path, NLPCR_URI_PATHSZ, 1);
	sv = path;

	while (9 != *p && 32 != *p && 0 != *p)
		*path++ = *p++;

	/* FIXME */
	/*if ( ! uri_pct_decode(sv, strlen(path))) {
		dddbg("%s: fail: allow-path pct-decode: %s",
				__func__, st->host);
		free(sv);
		return(0);
	} else*/
		dddbg("%s: allow-path: %s, %s", __func__,
				sv, st->host);

	assert(NULL != st->parser);
	assert(NULL != st->parser->allow);
	q_add(&st->parser->allow, sv);
	++(st->sz_allow);
	st->parser->fields |= REP_ALLOW;

	return(1);
}


static int
parse_deny(struct rstate_t *st)
{
	const char	*p;
	char		*path;
	size_t		i;

	assert(NULL != st);
	st->recording = 1;

	if (0 == st->active)
		return(1);
	if ( ! parse_skiptoval("Disallow", st, &p))
		return(0);
	assert(NULL != p);

	assert(NULL != st->parser);
	if (st->parser->max_deny < st->sz_deny + 1 &&
			0 != st->parser->max_deny) {
		dddbg("%s: fail: deny-path queue limit: %s",
				__func__, st->host);
		return(0);
	}

	calloc_s(path, NLPCR_URI_PATHSZ, 1);

	for (i = 0; 9 != p[i] && 32 != p[i] && 0 != p[i]; i++)
		path[i] = p[i];
	assert(i < NLPCR_URI_PATHSZ - 1);
	assert(i > 0);

	dddbg("%s: deny-path: %s, %s", __func__, path, st->host);

	assert(NULL != st->parser);
	assert(NULL != st->parser->deny);
	q_add(&st->parser->deny, path);
	++(st->sz_deny);
	st->parser->fields |= REP_DENY;

	return(1);
}


static int
parse_delay(struct rstate_t *st)
{
	int		t;
	const char	*p;

	assert(NULL != st);
	st->recording = 1;

	if (0 == st->active)
		return(1);
	if ( ! parse_skiptoval("Crawl-Delay", st, &p))
		return(0);

	assert(NULL != p);
	if ( ! parse_int(p, &t, INT_MAX, 0)) {
		dddbg("%s: fail: crawl-delay parse: %s",
				__func__, st->host);
		return(0);
	} else
		dddbg("%s: crawl-delay: %d, %s", 
				__func__, t, st->host);

	assert(NULL != st->parser);
	st->parser->crawl_delay = (uint32_t)t;
	st->parser->fields |= REP_DELAY;

	return(1);
}


static int
parse_line(struct rstate_t *st)
{
	size_t		sz;

	assert(NULL != st);
	assert(NULL != st->buf);

	if (0 == st->bufsz)
		return(1);
	if (32 == st->buf[0] || 9 == st->buf[0])
		return(1);

	if (st->bufsz >= (sz = strlen("User-Agent")) &&
			0 == strncasecmp(st->buf, "User-Agent", sz))
		return(parse_useragent(st));
	if (st->bufsz >= (sz = strlen("Disallow")) &&
			0 == strncasecmp(st->buf, "Disallow", sz))
		return(parse_deny(st));
	if (st->bufsz >= (sz = strlen("Allow")) && 
			0 == strncasecmp(st->buf, "Allow", sz))
		return(parse_allow(st));
	if (st->bufsz >= (sz = strlen("Crawl-Delay")) &&
			0 == strncasecmp(st->buf, "Crawl-Delay", sz))
		return(parse_delay(st));

	return(1);
}


int
rep_read(const char *host, int fd, char *buf)
{
	ssize_t		sz;

	assert(NULL != host);
	assert(NULL != buf);

	if (-1 == (sz = read(fd, buf, 1)))
		die("%s: fail: read: %s", __func__, host);
	else if (1 < sz)
		diex("%s: fail: read (overrun): %s", __func__, host);
	else if (0 == sz)
		return(0);
	return(1);
}


int
rep_parse(const char *host, int fd, struct rep_parse_t *res)
{
	char		buf[NLPCR_URI_PATHSZ];
	size_t		i;
	struct rstate_t	state;

	assert(NULL != host);
	assert(NULL != res);

	if (-1 == lseek(fd, 0, SEEK_SET))
		die("%s: fail: lseek: %s", __func__, host);
	
	memset(&state, 0, sizeof(struct rstate_t));

	state.buf = buf;
	state.host = host;
	state.parser = res;

	for (i = 0; i < NLPCR_URI_PATHSZ - 1; ) {
		if ( ! rep_read(host, fd, &buf[i]))
			break;

		/* 
		 * Line-endings.  Allow CR, LF, and CR/LF to act as line
		 * endings by treating them all equally (a CR/LF will
		 * thus be punctuated by a blank line).  
		 */
		if (13 != buf[i] && 10 != buf[i] && 35 != buf[i]) {
			i++;
			continue;
		} else if (35 == buf[i]) {
			i++;
			state.comment = 1;
			continue;
		} else if (0 == buf[i]) {
			dddbg("%s: null byte: ignoring line: %s", 
					__func__, host);
			i++;
			state.comment = 1;
		} else
			buf[i] = 0;

		if (state.comment) {
			state.comment = 0;
			i = 0;
			continue;
		}

		/*
		 * Once we've hit a "line", we push it into the main
		 * processor itself.  Zero all per-buffer metrics.
		 */

		state.bufsz = i;
		if ( ! parse_line(&state))
			return(0);
		i = 0;
	}

	if (NLPCR_URI_PATHSZ - 1 == i) {
		dddbg("%s: line too long: %s", __func__, host);
		return(0);
	}
	return(1);
}


void
rep_parse_init(size_t max_allow, size_t max_deny, 
		struct rep_parse_t *p)
{
	assert(NULL != p);
	memset(p, 0, sizeof(struct rep_parse_t));

	p->max_allow = max_allow;
	p->max_deny = max_deny;

	q_new(&p->allow);
	q_new(&p->deny);
}


void
rep_parse_free(struct rep_parse_t *p)
{
	char		*buf;

	assert(NULL != p);
	assert(NULL != p->allow);
	assert(NULL != p->deny);

	while (NULL != (buf = q_rem(&p->allow)))
		free(buf);
	while (NULL != (buf = q_rem(&p->deny)))
		free(buf);
	
	q_free(&p->allow);
	q_free(&p->deny);
}


