/* $Id: scan.c,v 1.104 2007-06-05 23:14:44 kristaps Exp $ */
/*
Copyright (c) 2007 Kristaps Dzonsons
Latvijas Universitaates Matemaatikas un Informaatikas Instituuts
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

* Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
* Neither the name of the copyright holders nor the names of
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/param.h>

#include <assert.h>
#include <err.h>
#include <ctype.h>
#include <getopt.h>
#include <inttypes.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include <libxml/HTMLparser.h>
#include <libxml/HTMLtree.h>

#include "nlpcrawl.h"


/*
 * These define various tags used during the parse sequence.  Since
 * they're all in the standard ASCII character domain, we can cast the
 * strings directly into xmlChar pointers.
 */
#define XML_ATTR_HREF	((xmlChar *)"href")
#define XML_ATTR_NAME	((xmlChar *)"name")
#define XML_ATTR_CONTENT ((xmlChar *)"content")
#define XML_ATTR_EQUIV	((xmlChar *)"http-equiv")
#define XML_TAG_A	((xmlChar *)"a")
#define XML_TAG_LINK	((xmlChar *)"link")
#define XML_TAG_BASE	((xmlChar *)"base")
#define XML_TAG_META	((xmlChar *)"meta")
#define XML_TAG_NIDEX	((xmlChar *)"noindex")

/*
 * These are for the Robots Exclusion Standard (which define the index
 * and follow tokens, and respective no- prefix).  If index is not
 * defined in a bitmask, we drop the page from the database (via going
 * to UNREACH).  If follow isn't defined, do not process the page for
 * links.
 */
#define SCAN_NOINDEX	(1 << 0)
#define SCAN_NOFOLLOW	(1 << 1)


static struct opt_t {
	void		*scheme;	/* Htable of schemes.        */
	void		*auth;		/* Htable of authorities.    */
	void		*queue;		/* Write-queue.              */
	char		*cache;		/* Cache-file directory.     */
	int		ascend;		/* Perform path-ascendence.  */
	int		sascend;	/* Path ascend only vrfy.    */
} opts;



static void
proc_add(struct uri *addr)
{
	struct nlpcdbrp	*n;

	assert(NULL != addr);

	calloc_s(n, 1, sizeof(struct nlpcdbrp));
	nlpcr_new(addr, n);
	q_add(&opts.queue, n);
}


static void
proc_link(const char *link, const struct uri *base)
{
	struct uri	p, addr;
	char		buf[NLPCR_URI_FMTSZ], *auth, *sch;

	assert(NULL != link);

	if (NULL != base) {
		if ( ! uri_parse(link, 0, &p)) {
			ddbg("%s: fail: bad uri: %s", __func__, link);
			return;
		} else if ( ! uri_ref_resolve(base, &p, &addr)) {
			ddbg("%s: fail: ref-res: %s", __func__, link);
			return;
		}
	} else if ( ! uri_parse(link, 0, &addr)) {
		ddbg("%s: fail: bad uri: %s", __func__, link);
		return;
	}

	uri_topstring_r(&addr, buf);
	memset(addr.fragment, 0, NLPCR_URI_FRAGSZ);
	auth = addr.auth;
	sch = addr.scheme;

	if ( ! uri_is_abs(&addr))
		ddbg("%s: fail: rel. uri: %s", __func__, buf);
	else if (0 == auth[0])
		ddbg("%s: fail: no auth: %s", __func__, buf);
	else if (0 == sch[0])
		ddbg("%s: fail: no scheme: %s", __func__, buf);
	else if ( ! dict_has(&opts.scheme, sch, strlen(sch)))
		ddbg("%s: fail: scheme miss: %s", __func__, buf);
	else if ( ! dict_has(&opts.auth, auth, strlen(auth)))
		ddbg("%s: reject: auth miss: %s", __func__, buf);
	else if (opts.ascend) 
		while (uri_ascend(&addr))
			proc_add(&addr);
	else
		proc_add(&addr);
}


static void
proc_robot(xmlNodePtr ptr, unsigned int *flags)
{
	xmlChar		*ref;

	assert(NULL != ptr);
	assert(NULL != flags);

	if (NULL == (ref = xmlGetProp(ptr, XML_ATTR_CONTENT)))
		return;

	*flags = 0;
	if (NULL != xmlStrcasestr(ref, (xmlChar *)"all")) 
		*flags = 0;
	if (NULL != xmlStrcasestr(ref, (xmlChar *)"none"))
		*flags = SCAN_NOFOLLOW | SCAN_NOINDEX;
	if (NULL != xmlStrcasestr(ref, (xmlChar *)"index"))
		*flags &= ~SCAN_NOINDEX;
	if (NULL != xmlStrcasestr(ref, (xmlChar *)"noindex"))
		*flags |= SCAN_NOINDEX;
	if (NULL != xmlStrcasestr(ref, (xmlChar *)"follow"))
		*flags &= ~SCAN_NOFOLLOW;
	if (NULL != xmlStrcasestr(ref, (xmlChar *)"nofollow"))
		*flags |= SCAN_NOFOLLOW;

	if (SCAN_NOFOLLOW & *flags)
		ddbg("%s: scan no-follow raised", __func__);
	if (SCAN_NOINDEX & *flags)
		ddbg("%s: scan no-index raised", __func__);

	free(ref);
}


static void
proc_meta(xmlNodePtr ptr, unsigned int *flags)
{
	xmlChar		*ref;

	assert(NULL != ptr);
	assert(NULL != flags);

	if (NULL != (ref = xmlGetProp(ptr, XML_ATTR_NAME))) {
		if (0 == xmlStrcasecmp(ref, (xmlChar *)"robots"))
			proc_robot(ptr, flags);
	} else
		return;

	free(ref);
}


static void
proc_anchor(xmlNodePtr ptr, struct uri *base)
{
	xmlChar		*ref;

	assert(NULL != ptr);
	assert(NULL != base);

	if (NULL == (ref = xmlGetProp(ptr, XML_ATTR_HREF)))
		return;
	proc_link((char *)ref, base);
	xmlFree(ref);
}


static void
proc_base(xmlNodePtr ptr, struct uri *baseuri)
{
	struct uri 	p;
	xmlChar		*ref;
	char		*buf;

	assert(NULL != baseuri);
	assert(NULL != ptr);

	if (NULL == (ref = xmlGetProp(ptr, XML_ATTR_HREF)))
		return;
	buf = (char *)ref;

	if (uri_parse(buf, 0, &p)) {
		if ( ! uri_is_abs(&p))
			ddbg("%s: warn: uri not absolute: %s",
					__func__, buf);
		else	memcpy(baseuri, &p, sizeof(struct uri));
	} else
		ddbg("%s: warn: uri failed parse: %s", __func__, buf);

	xmlFree(ref);
}


static void
proc_node(struct uri *base, xmlNodePtr node, unsigned int *flags)
{
	xmlNodePtr 	c;
	const xmlChar	*p;

	assert(NULL != base);

	for (c = node; NULL != c; c = c->next) {
		if (c->type == XML_ELEMENT_NODE) {
			p = c->name;
			assert(NULL != p);
			if (0 == xmlStrcasecmp(XML_TAG_BASE, p))
				proc_base(c, base);
			else if (0 == xmlStrcasecmp(XML_TAG_META, p))
				proc_meta(c, flags);
			else if (0 == xmlStrcasecmp(XML_TAG_A, p))
				proc_anchor(c, base);
			else if (0 == xmlStrcasecmp(XML_TAG_LINK, p))
				proc_anchor(c, base);
			else if (0 == xmlStrcasecmp(XML_TAG_NIDEX, p))
				if (SCAN_NOFOLLOW & *flags)
					return;
		}
		if (SCAN_NOFOLLOW & *flags)
			return;
		proc_node(base, c->children, flags);
		if (SCAN_NOFOLLOW & *flags)
			return;
	}
}


static int
proc_html(struct nlpcdbrp *pair, unsigned int *flags)
{
	char		buf[NLPCR_URI_FMTSZ],
			path[MAXPATHLEN];
	struct uri	baseuri;
	struct nlpcdbr	*r;
	htmlDocPtr 	p;
	xmlNodePtr 	n;

	assert(NULL != pair);
	assert(NULL != flags);

	r = &pair->record;
	uri_topstring_r(&r->uri, buf);
	nlpcdbr_hash2path_r(path, opts.cache, pair->hash);

	assert(0 != r->charset[0]);
	if (NULL == (p = htmlParseFile(path, r->charset))) {
		r->code = NLPCDBR_SYNTAX;
		ddbg("%s: fail: parse: %s", __func__, buf);
		return(0);
	}

	memcpy(&baseuri, &r->uri, sizeof(struct uri));
	if (NULL == (n = xmlDocGetRootElement(p))) {
		r->code = NLPCDBR_SYNTAX;
		ddbg("%s: fail: acquire root: %s", __func__, buf);
		xmlFreeDoc(p);
		return(0);
	}

	proc_node(&baseuri, n, flags);
	xmlFreeDoc(p);
	return(1);
}


static int
proc(struct nlpcdbrp *pair)
{
	unsigned int 	flags;

	assert(NULL != pair);

	flags = 0;

	switch (pair->record.type) {
	case (NLPCDBR_TYPE_TXT_HTML):
		if ( ! proc_html(pair, &flags))
			return(-1);
		if (SCAN_NOINDEX & flags) {
			pair->record.code = NLPCDBR_REP;
			return(-1);
		}
		break;
	case (NLPCDBR_TYPE_TXT_PLAIN):
		break;
	default:
		ddbg("warn: unsupported type: %s, %s", 
				nlpcdbr_type2str(pair->record.type), 
				uri_topstring(&pair->record.uri));
		return(0);
	}

	return(1);
}


static int
scan_queue(void)
{
	char		buf[NLPCR_URI_FMTSZ], reason[32];
	int		s, u, n, flags;
	struct nlpcdbrp	*p;

	if (q_empty(&opts.queue))
		return(0);

	for (s = 0, n = 0, u = 0, flags = 0; ; ) {
		if (NULL == (p = q_rem(&opts.queue)))
			break;

		uri_topstring_r(&p->record.uri, buf);
		nlpcdbr_code2str_r(reason, p->record.code);

		if (NLPCDBR_NEW == p->record.state) {
			if ( ! db_add(p)) {
				ddbg("reject: duplicate: %s", buf);
				goto scan_queue_tail;
			}
		} else if ( ! db_put(p))
			diex("%s: fail: db_put: %s", __func__, buf);

		switch (p->record.state) {
		case (NLPCDBR_SCAN):
			dbg("accept: scanned: %s", buf);
			s++;
			break;
		case (NLPCDBR_NEW):
			dbg("accept: new: %s", buf);
			n++;
			break;
		case (NLPCDBR_UNREACH):
			dbg("accept: unreach (%s): %s", reason, buf);
			u++;
			break;
		default:
			abort();
		}

scan_queue_tail:
		free(p);
		if (SIGNAL_EXIT & (flags |= signal_test()))
			break;
	}
	if (n > 0)
		signal_next(NLPCDBR_NEW);
	if (u > 0)
		signal_next(NLPCDBR_UNREACH);
	if (s > 0)
		signal_next(NLPCDBR_SCAN);
	return(flags);
}


static int
scan(void)
{
	void		*cookie;
	struct db_query	query;
	int		rc, flags;
	struct nlpcdbrp *p;
	struct nlpcdbr	*r;

	memset(&query, 0, sizeof(struct db_query));

	query.field = DB_FIELD_STATE;
	query.state = NLPCDBR_VRFY;

	db_window_fill(&query);
	db_query_begin(&cookie);

	for (flags = 0, p = NULL; ; ) {
		if (SIGNAL_EXIT & (flags |= signal_test()))
			break;
		if (NULL == p)
			calloc_s(p, 1, sizeof(struct nlpcdbrp));
		if (1 != db_query_next(&cookie, &query, p))
			break;
		if (0 == (rc = proc(p)))
			continue;

		r = &p->record;
		r->atime = r->mtime = time_uint64();

		if (1 == rc) {
			r->state = NLPCDBR_SCAN;
			r->revived = 0;
		} else
			r->state = NLPCDBR_UNREACH;

		q_add(&opts.queue, p);
		if (NLPCDBR_SCAN == r->state && opts.sascend)
			while (uri_ascend(&r->uri))
				proc_add(&r->uri);
		p = NULL;
	}
	db_query_close(&cookie);
	if (NULL != p)
		free(p);

	if (SIGNAL_EXIT & flags)
		return(flags);
	return(flags | scan_queue());
}


static int
scan_seed(int argc, char *argv[])
{
	char		*p;

	assert(NULL != argv);
	while (argc-- > 0 && NULL != (p = *argv++))
		proc_link(p, NULL);
	return(scan_queue());
}


void
xml_struct(void *arg, xmlErrorPtr p)
{
	(void)arg;
	(void)p;
}


void
xml_generic(void *arg, const char *msg, ...)
{
	(void)arg;
	(void)msg;
}


int
main(int argc, char *argv[])
{
	int 		c;
	char		*dir;

	extern char	*__progname;
	extern char 	*optarg;
	extern int 	optind;
	struct nlpcdbrp	*p;

	static struct option longopts[] = {
		{ "filter-scheme", 	required_argument, NULL, 1 },
		{ "filter-auth", 	required_argument, NULL, 2 },
		{ "filter-ascend", 	no_argument, NULL, 3 },
		{ "filter-ascend-scan", no_argument, NULL, 4 },
		{ NULL, 0, NULL, 0 }
	};

	dir = NLPC_DEFAULT_DIR;

	memset(&opts, 0, sizeof(struct opt_t));
	opts.cache = NLPC_DEFAULT_FILES;

	q_new(&opts.queue);

	while (-1 != (c = getopt_long(argc, argv, "df:hp:v",
					longopts, NULL)))
		switch (c) {
		case (1):
			dict_free(&opts.scheme);
			dict_fill(&opts.scheme, optarg);
			break;
		case (2):
			dict_free(&opts.auth);
			dict_fill(&opts.auth, optarg);
			break;
		case (3):
			opts.ascend = 1;
			break;
		case (4):
			opts.sascend = 1;
			break;
		case ('d'):
			dbg_lvl++;
			break;
		case ('f'):
			opts.cache = optarg;
			break;
		case ('p'):
			dir = optarg;
			break;
		case ('v'):
			fprintf(stdout, "%s version %s\n", 
					__progname, VERSION);
			return(EXIT_SUCCESS);
		case ('h'):
			/*FALLTHROUGH*/
		default:
			fprintf(stderr, "usage: %s [-dhv] [-f path] "
					"[-p path] [LONG_OPTS] "
					"[uri ...]\n", __progname);
			return(EXIT_FAILURE);
		}
	argc -= optind;
	argv += optind;

	xmlSetStructuredErrorFunc(NULL, xml_struct);
	xmlSetGenericErrorFunc(NULL, xml_generic);

	signal_init(dir, NLPCDBR_VRFY);
	db_init(dir, 0);

	if (SIGNAL_EXIT & scan_seed(argc, argv))
		goto main_out;

	for (;;) {
		if (SIGNAL_EXIT & scan())
			break;
		if (SIGNAL_EXIT & signal_wait(0))
			break;
	}

main_out:

	dict_free(&opts.scheme);
	dict_free(&opts.auth);

	while (NULL != (p = q_rem(&opts.queue)))
		free(p);
	q_free(&opts.queue);

	db_teardown();
	xmlCleanupParser();
	signal_teardown();

	return(EXIT_SUCCESS);
}

